import logging

from goblet.client import VersionedClients, get_default_location, get_default_project

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.INFO)


class Handler:
    """Base Handler class"""

    valid_backends = []
    resources = None
    resource_type = ""
    backend = "cloudfunction"
    can_sync = False

    def __init__(
        self,
        name,
        versioned_clients: VersionedClients = None,
        resources=None,
        backend="cloudfunction",
    ):
        self.name = name
        self.backend = backend
        self.resources = resources or {}
        self.versioned_clients = versioned_clients or VersionedClients()
        self.cloudfunction = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{name}"

    def deploy(self, source=None, entrypoint=None, config={}):
        if self.resources and self.backend not in self.valid_backends:
            log.info(
                f"skipping... {self.backend} not supported for {self.resource_type}"
            )
            return
        self._deploy(source, entrypoint, config=config)

    def _deploy(self, source=None, entrypoint=None, config={}):
        raise NotImplementedError("deploy")

    def destroy(self):
        raise NotImplementedError("destroy")

    def sync(self, dryrun=False):
        if self.can_sync:
            log.info(f"syncing {self.resource_type}")
            self._sync(dryrun)

    def _sync(self, dryrun=False):
        pass

    def __call__(self, request, context=None):
        raise NotImplementedError("__call__")

    def __add__(self, other):
        if other.resources:
            self.resources.update(other.resources)
        return self



import bz2
import json
import re

from os import listdir
from os.path import join, exists, dirname

from httplib2 import Http, Response
from six.moves.urllib.parse import urlparse

DATA_DIR = join(dirname(__file__), "tests", "data", "http")
PROJECT_ID = "goblet"
DATA_DIR_MAIN = join(dirname(__file__), "tests", "data")


def sanitize_project_name(dirty_str):
    sanitized = "projects/{}/".format(PROJECT_ID)
    return re.sub(r"projects/([0-9a-zA-Z_-]+)/", sanitized, dirty_str)


def dummy_function():
    return None


def mock_dummy_function(mock):
    def dummy_function(event=None):
        return mock()

    return dummy_function


class HttpFiles(Http):
    """Logic to interate through files in the DATA_DIR for both recording and replaying http responses"""

    def __init__(self, data_path, discovery_path):
        self._data_path = data_path
        self._discovery_path = discovery_path
        self._index = {}
        super(HttpFiles, self).__init__()

    def get_next_file_path(self, uri, method, record=True):
        uri = sanitize_project_name(uri)
        base_name = "%s%s" % (
            method.lower(),
            urlparse(uri).path.replace("/", "-").replace(":", "-"),
        )
        data_dir = self._data_path

        is_discovery = False
        # We don't record authentication
        if (
            base_name.startswith("post-oauth2-v4")
            or base_name.startswith("post-o-oauth2")
            or base_name.startswith("post-token")
        ):
            return
        # Use a common directory for discovery metadata across tests.
        if base_name.startswith("get-$discovery"):
            base_name = base_name.replace("-$d", "-d")
        if base_name.startswith("get-discovery"):
            data_dir = self._discovery_path
            is_discovery = True

        next_file = None
        while next_file is None:
            index = self._index.setdefault(base_name, 1)
            fn = join(data_dir, "{}_{}.json".format(base_name, index))
            if is_discovery:
                fn += ".bz2"
            if exists(fn):
                # if we already have discovery metadata, don't re-record it.
                if record and is_discovery:
                    return None
                # on replay always return the same discovery file
                if is_discovery:
                    return fn
                self._index[base_name] += 1
                if not record:
                    next_file = fn
            elif record:
                return fn
            else:
                raise IOError("response file ({0}) not found".format(fn))

        return fn


class HttpRecorder(HttpFiles):
    """Save the output of http responses for future testing"""

    def request(
        self,
        uri,
        method="GET",
        body=None,
        headers=None,
        redirections=1,
        connection_type=None,
    ):
        response, content = super(HttpRecorder, self).request(
            uri, method, body, headers, redirections, connection_type
        )
        fpath = self.get_next_file_path(uri, method)

        if fpath is None:
            return response, content

        fopen = open
        if fpath.endswith(".bz2"):
            fopen = bz2.BZ2File
        with fopen(fpath, "wb") as fh:
            recorded = {}
            recorded["headers"] = {}
            if not content:
                content = "{}"
            recorded["body"] = json.loads(content)
            fh.write(
                sanitize_project_name(json.dumps(recorded, indent=2)).encode("utf8")
            )

        return response, content


class HttpReplay(HttpFiles):
    """Replay the output of http responses for tests"""

    static_responses = {
        ("POST", "https://accounts.google.com/o/oauth2/token"): json.dumps(
            {"access_token": "ya29", "token_type": "Bearer", "expires_in": 3600}
        ).encode("utf8"),
        ("POST", "https://oauth2.googleapis.com/token"): json.dumps(
            {"access_token": "ya29", "token_type": "Bearer", "expires_in": 3600}
        ).encode("utf8"),
    }

    _cache = {}

    def request(
        self,
        uri,
        method="GET",
        body=None,
        headers=None,
        redirections=1,
        connection_type=None,
    ):
        if (method, uri) in self.static_responses:
            return (
                Response(
                    {"status": "200", "content-type": "application/json; charset=UTF-8"}
                ),
                self.static_responses[(method, uri)],
            )

        fpath = self.get_next_file_path(uri, method, record=False)
        fopen = open
        if fpath.endswith(".bz2"):
            if fpath in self._cache:
                return self._cache[fpath]
            fopen = bz2.BZ2File
        with fopen(fpath, "rb") as fh:
            data = json.load(fh)
            response = Response(data["headers"])
            serialized = json.dumps(data["body"]).encode("utf8")
            if fpath.endswith("bz2"):
                self._cache[fpath] = response, serialized
            return response, serialized


def get_responses(folder):
    """Get responses from the DATA_DIR for validating tests"""
    responses = []
    for response in sorted(listdir(f"{DATA_DIR}/{folder}")):
        with open(f"{DATA_DIR}/{folder}/{response}") as f:
            responses.append(json.loads(f.read()))

    return responses


def get_response(folder, filename):
    """Get a response from the DATA_DIR for validating tests"""
    with open(f"{DATA_DIR}/{folder}/{filename}") as f:
        return json.loads(f.read())



import json
from goblet.utils import get_g_dir, nested_update
import os
import logging

log = logging.getLogger("goblet.config")
log.setLevel(logging.INFO)


class GConfig:
    """Config class used to get variables from config.json or from the environment. If stage is set as an environment level
    if will parse the corresponding section in config.json and return those config values"""

    def __init__(self, config=None, stage=None):
        self.config = self.get_g_config()
        if config:
            self.config = nested_update(self.config, config)
        self.stage = stage or os.environ.get("STAGE")
        self.validate()
        if self.stage:
            self.config = nested_update(
                self.config, self.config.get("stages", {}).get(self.stage, {})
            )

    @staticmethod
    def get_g_config():
        try:
            with open(f"{get_g_dir()}/config.json") as f:
                return json.load(f)
        except FileNotFoundError:
            return {}
        except json.decoder.JSONDecodeError:
            log.info(
                "JSONDecodeError. config.json is not valid. Returning empty config"
            )
            return {}

    def __getattr__(self, name):
        if os.environ.get(name):
            return os.environ.get(name)
        attr = self.config.get(name)
        if attr:
            return attr
        return None

    def __setattr__(self, name, value):
        if name not in ["config", "stage"]:
            self.config[name] = value
        else:
            super(GConfig, self).__setattr__(name, value)

    def write(self):
        with open(f"{get_g_dir()}/config.json", "w") as f:
            f.write(json.dumps(self.config, indent=4))

    def validate(self):
        if self.stage and self.stage not in self.config.get("stages"):
            raise ValueError(f"stage {self.stage} not found in config")
        for stage in self.config.get("stages", {}):
            if "function_name" not in self.config["stages"][stage]:
                raise ValueError(f"function_name key missing for stage {stage}")



import base64
import hashlib
import logging
import os
import warnings
import zipfile
from pathlib import Path

import requests
from googleapiclient.errors import HttpError

from goblet.config import GConfig
from goblet.utils import get_g_dir, checksum


class Backend:
    """Base backend class"""

    resource_type = ""
    version = ""

    def __init__(self, app, client, func_path, config={}, zip_config=None):
        self.app = app
        self.name = app.function_name
        self.log = logging.getLogger("goblet.backend")
        self.log.setLevel(logging.INFO)
        self.zip_path = get_g_dir() + f"/{self.name}.zip"
        self.zipf = self._create_zip()
        self.config = GConfig(config=config)

        # specifies which files to be zipped
        self.zip_config = zip_config or {
            "include": ["*.py"],
            "exclude": ["build", "docs", "examples", "test", "tests", "venv"],
        }

        self.func_path = func_path

        self.client = client

    def deploy(self, force=False, config=None):
        raise NotImplementedError("destroy")

    def destroy(self, all=False):
        raise NotImplementedError("destroy")

    def _create_zip(self):
        """Creates initial goblet zipfile"""
        if not os.path.isdir(get_g_dir()):
            os.mkdir(get_g_dir())
        return zipfile.ZipFile(self.zip_path, "w", zipfile.ZIP_DEFLATED)

    def delta(self, zip_path=None):
        """Compares md5 hash between local zipfile and cloudfunction already deployed"""
        if zip_path is None:
            zip_path = self.zip_path
        self.zipf.close()
        with open(zip_path, "rb") as fh:
            local_checksum = base64.b64encode(checksum(fh, hashlib.md5())).decode(
                "ascii"
            )

        deployed_checksum = self._checksum()
        modified = deployed_checksum != local_checksum
        return modified

    def _checksum(self):
        raise NotImplementedError("_checksum")

    def _gcs_upload(self, client, headers, upload_client=None, force=False):
        self.log.info("zipping source code")
        self.zip()
        if not force and self.get() and not self.delta():
            self.log.info("No changes detected....")
            return None, False
        self.log.info("uploading source zip to gs......")
        return self._upload_zip(upload_client or client, headers), True

    def _upload_zip(self, client, headers=None) -> dict:
        """Uploads zipped cloudfunction using generateUploadUrl endpoint"""
        self.zipf.close()
        with open(f".goblet/{self.name}.zip", "rb") as f:
            resp = client.execute("generateUploadUrl", params={"body": {}})
            try:
                requests.put(
                    resp["uploadUrl"],
                    data=f,
                    headers=headers,
                ).raise_for_status()
            except requests.exceptions.HTTPError as e:
                if not os.environ.get("GOBLET_HTTP_TEST") == "REPLAY":
                    raise e

        self.log.info("source code uploaded")

        return resp

    def get(self):
        """Returns backend currently deployed or None"""
        try:
            return self.client.execute(
                "get", parent_key="name", parent_schema=self.func_path
            )
        except HttpError as e:
            if e.resp.status != 404:
                raise

    def zip(self):
        """Zips requirements.txt, python files and any additional files based on config.customFiles"""
        self._zip_file("requirements.txt")
        if self.config.main_file:
            self._zip_file(self.config.main_file, "main.py")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._zip_directory()

    def _zip_file(self, filename, arcname=None):
        self.zipf.write(filename, arcname)

    def _zip_directory(self):
        exclusion_set = set(self.zip_config.get("exclude", []))
        globbed_files = []
        for pattern in self.zip_config.get("include", []):
            globbed_files.extend(Path("").rglob(pattern))
        for path in globbed_files:
            if not set(path.parts).intersection(exclusion_set):
                self.zipf.write(str(path))
